/*

Modified from code copyright (C) 2006 Pedro Felzenszwalb, under the terms of the GNU General Public License as published by
the Free Software Foundation.

*/

#ifndef SEGMENT_IMAGE
#define SEGMENT_IMAGE

#include <cstdlib>
//#include <image.h>
#include <algorithm>
#include <cmath>
#include "segment-graph.h"

typedef struct {
  float val;
  int ind, loc;
} ind_val;

bool operator<(const ind_val &a, const ind_val &b) {
  return a.val > b.val;
}

static inline float diff2(float *r, int x1, int t1, int y1, int x2, int t2, int y2, int n1, int n2, float e1, float e2) {
  float dist=sqrt((x1-x2)*(x1-x2) + (t1-t2)*(t1-t2) + (y1-y2)*(y1-y2)) / e1;
  float amp=r[t2+n1*x2+y2*n1*n2] / e2;
  return exp((dist+amp));
}



/*
 * Segment an image
 *
 * Returns a color image representing the segmentation.
 *
 * im: image to segment.
 * sigma: to smooth the image.
 * c: constant for treshold function.
 * min_size: minimum component size (enforced by post-processing stage).
 * num_ccs: number of connected components in the segmentation.
 */
//image<rgb> *segment_image(image<rgb> *im, float sigma, float c, int min_size, int *num_ccs) {

float *segment_image(float *im, int n1, int n2, int n3, float sigma, float c, int min_size, float e1, float e2, int max_dist, int *num_ccs) {
  int width = n2, ck, ckx, x1, x2, t1, t2, y1, y2;
  int height = n1;
  int depth = n3;
  int hw = height*width;
  float *output; 
  

fprintf(stderr,"in segment_image: %d x %d \n",width,height);

  // build graph
  edge *edges = new edge[width*height*depth*(max_dist)*10];


  int num = 0;
 for (int y1 = 0; y1 < depth; y1++)  {
fprintf(stderr,"slice %d \n",y1);
  for (int t1 = 0; t1 < height; t1++) {
    for (int x1 = 0; x1 < width; x1++) {

    int cx=0, cy=0, cxy=0, cxy2=0, cz=0, czx=0, czy=0, czxy=0, czxy2=0, czxy3=0, czx2=0, czy2=0;

 
   ind_val *pathx = new ind_val[max_dist];
   ind_val *pathy = new ind_val[max_dist];
   ind_val *pathxy = new ind_val[max_dist];
   ind_val *pathxy2 = new ind_val[max_dist];
   ind_val *pathz = new ind_val[max_dist];
   ind_val *pathzx = new ind_val[max_dist];
   ind_val *pathzx2 = new ind_val[max_dist];
   ind_val *pathzy = new ind_val[max_dist];
   ind_val *pathzy2 = new ind_val[max_dist];
   ind_val *pathzxy = new ind_val[max_dist];
   ind_val *pathzxy2 = new ind_val[max_dist];
   ind_val *pathzxy3 = new ind_val[max_dist];

       for (int px=0; px < max_dist; px++) {
          int x2 = x1 + px +1;
          int t2 = t1 + px +1;
          int t3 = t1 - (px+1);
          int y2 = y1 + px +1;
          int y3 = y1 - (px+1);

          if (x2 < width) {
             pathx[px].ind = px;
             pathx[px].val = im[y1 * hw + t1 * width + x2];
             pathx[px].loc = y1 * hw + t1 * width + x2;
             cx++;
           }
          if (t2 < height) {
             pathy[px].ind = px;
             pathy[px].val = im[y1 * hw + t2 * width + x1];
             pathy[px].loc = y1 * hw + t2 * width + x1;
             cy++;
           }
           if (y2 < depth) {
             pathz[px].ind = px;
             pathz[px].val = im[y2 * hw + t1 * width + x1];
             pathz[px].loc = y2 * hw + t1 * width + x1;                        
             cz++;
           }
           
         if ((x2 < width) && (t2 < height)) {
             pathxy[px].ind = px;
             pathxy[px].val = im[y1 * hw + t2 * width + x2];
             pathxy[px].loc = y1 * hw + t2 * width + x2;
             cxy++;
           }
         if ( (y2 < depth) && (t2 < height) && (x2 < width) ) {
             pathzxy[px].ind = px;
             pathzxy[px].val = im[y2 * hw + t2 * width + x2];
             pathzxy[px].loc = y2 * hw + t2 * width + x2;                        
             czxy++;
           }           
         if ((x2 < width) && (t3 > 0) && (t2 < height)) {
             pathxy2[px].ind = px;
             pathxy2[px].val = im[y1 * hw + t3 * width + x2];
             pathxy2[px].loc = y1 * hw + t3 * width + x2;                        
             cxy2++;
           }

         if ( (y2 < depth) && (x2 < width) ) {
             pathzx[px].ind = px;
             pathzx[px].val = im[y2 * hw + t1 * width + x2];
             pathzx[px].loc = y2 * hw + t1 * width + x2;                        
             czx++;
           }
         if ( (y2 < depth) && (t2 < height) ) {
             pathzy[px].ind = px;
             pathzy[px].val = im[y2 * hw + t2 * width + x1];
             pathzy[px].loc = y2 * hw + t2 * width + x1;                        
             czy++;
           }


         if ( (y2 < depth) && (t2 < height) && (x2 < width) ) {
             pathzxy2[px].ind = px;
             pathzxy2[px].val = im[y2 * hw + t3 * width + x2];
             pathzxy2[px].loc = y2 * hw + t3 * width + x2;                        
             czxy2++;
           }

         if ((x2 < width) && (y3 > 0) && (t2 < height)) {
             pathzxy3[px].ind = px;
             pathzxy3[px].val = im[y3 * hw + t2 * width + x2];
             pathzxy3[px].loc = y3 * hw + t2 * width + x2;                        
             czxy3++;
           }

         if ( (y3 > 0) && (x2 < width) ) {
             pathzx2[px].ind = px;
             pathzx2[px].val = im[y3 * hw + t1 * width + x2];
             pathzx2[px].loc = y3 * hw + t1 * width + x2;                        
             czx2++;
          }

         if ( (y3 > 0) && (t2 < width) ) {
             pathzy2[px].ind = px;
             pathzy2[px].val = im[y3 * hw + t2 * width + x1];
             pathzy2[px].loc = y3 * hw + t2 * width + x1;                        
             czy2++;
          }

        std::sort(pathx, pathx + cx);
        std::sort(pathy, pathy + cy);
        std::sort(pathxy, pathxy + cxy);
        std::sort(pathxy2, pathxy2 + cxy2);
        std::sort(pathz, pathz + cz);
        std::sort(pathzx, pathzx + czx);        
        std::sort(pathzy, pathzy + czy);        
        std::sort(pathzxy, pathzxy + czxy);        
        std::sort(pathzxy2, pathzxy2 + czxy2);  
        std::sort(pathzxy3, pathzxy3 + czxy3);     
        std::sort(pathzx2, pathzx2 + czx2);
        std::sort(pathzy2, pathzy2 + czy2);                                 
      }

  


         if (cx > 0) { // Do pathx
           x2 = x1 + cx;				// Set other endpoint 
           ck = 0;
           for (int ix=0; ck == 0; ix++)  {
             if (pathx[ix].ind <= cx) {
                ck = 1;
                x2 = x1 + ix + 1;
                for (int iw=0; iw<cx; iw++) {
                edges[num+iw].a = y1 * hw + t1 * width + x1;
                edges[num+iw].b = y1 * hw + t1 * width + x2;
                edges[num+iw].w = diff2(im, x1, t1, y1, x2, t1, y1, n1, n2, e1, e2);
                } 
             }
            }
          
          num=num+cx; 
         }       

        if (cy > 0) {// Do pathy
           t2 = t1 + cy;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y1 * hw + t2 * width + x1;

           ck = 0;
           for (int iy=0; ck == 0; iy++)  {
             if (pathy[iy].ind <= cy) {
                ck = 1;
                t2 = t1 + iy + 1;
                edges[num].w = diff2(im, x1, t1, y1, x1, t2, y1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathy[iy].loc]));
             }
            }

           num++;        
         }
         
         if (cz > 0) { // Do pathz
           y2 = y1 + cz;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t1 * width + x1;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {
             if (pathz[iz].ind <= cz) {
                ck = 1;
                y2 = y1 + iz + 1;
                edges[num].w = diff2(im, x1, t1, y1, x1, t1, y2, n1, n2, e1, e2); 
             }
            }
           num++;        
          }    

       if (cxy > 0) { // Do pathxy
           x2 = x1 + cxy;
           t2 = t1 + cxy;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y1 * hw + t2 * width + x2;

           ck = 0;
           for (int id=0; ck == 0; id++)  {
             if (pathxy[id].ind <= cxy) {
                ck = 1;
                x2 = x1 + id + 1;
                t2 = t1 + id + 1;
                edges[num].w = diff2(im, x1, t1, y1, x2, t2, y1, n1, n2, e1, e2);  //sqrt(square(im[t1 * width + x1]-im[pathxy[id].loc]));
             }
            }

           num++;
           }     

         if (czxy > 0) { // Do pathzxy
           y2 = y1 + czxy;
           t2 = t1 + czxy;
           x2 = x1 + czxy;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t2 * width + x2;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {
             if (pathzxy[iz].ind <= czxy) {
                ck = 1;
                y2 = y1 + iz + 1;
                t2 = t1 + iz + 1;
                x2 = x1 + iz + 1;
                edges[num].w = diff2(im, x1, t1, y1, x2, t2, y2, n1, n2, e1, e2); 
             }
            }
           num++;
           }        
         
         if (czx > 0) { // Do pathzx
           y2 = y1 + czx;
           x2 = x1 + czx;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t1 * width + x2;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {
             if (pathzx[iz].ind <= czx) {
                ck = 1;
                y2 = y1 + iz + 1;
                x2 = x1 + iz + 1;
                edges[num].w = diff2(im, x1, t1, y1, x2, t1, y2, n1, n2, e1, e2); 
             }
            }
           num++; 
          }       
         
       if (cxy2 > 0) { // Do pathxy2
           x2 = x1 + cxy2;
           t2 = t1 - cxy2;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y1 * hw + t2 * width + x2;

           ck = 0;
           for (int id=0; ck == 0; id++)  {
             if (pathxy2[id].ind <= cxy2) {
                ck = 1;
                x2 = x1 + id + 1;
                t2 = t1 - (id + 1);
                edges[num].w = diff2(im, x1, t1, y1, x2, t2, y1, n1, n2, e1, e2);  
             }
            }
          num++;
         }

         if (czy > 0) { // Do pathzy
           y2 = y1 + czy;
           t2 = t1 + czy;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t2 * width + x1;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {
             if (pathzy[iz].ind <= czy) {
                ck = 1;
                y2 = y1 + iz + 1;
                t2 = t1 + iz + 1;
                edges[num].w = diff2(im, x1, t1, y1, x1, t2, y2, n1, n2, e1, e2); 
             }
            }
           num++;
           }        
          
         if (czxy2 > 0) { // Do pathzxy2
           y2 = y1 + czxy2;
           t2 = t1 - czxy2;
           x2 = x1 + czxy2;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t2 * width + x2;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {
             if (pathzxy2[iz].ind <= czxy2) {
                ck = 1;
                y2 = y1 + iz + 1;
                t2 = t1 + iz + 1;
                x2 = x1 + iz + 1;
                edges[num].w = diff2(im, x1, t1, y1, x2, t2, y2, n1, n2, e1, e2); 
             }
            }
           num++;  
           }      
       
         if (czxy3 > 0) { // Do pathzxy3
           y2 = y1 - czxy3;
           t2 = t1 + czxy3;
           x2 = x1 + czxy3;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t2 * width + x2;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {

             if (pathzxy3[iz].ind <= czxy3) {
                ck = 1;
                y2 = y1 - (iz + 1);
                t2 = t1 + iz + 1;
                x2 = x1 + iz + 1;

                edges[num].w = diff2(im, x1, t1, y1, x2, t2, y2, n1, n2, e1, e2); 
             }
            }
           num++;
           }        
      
         if (czy2 > 0) { // Do pathzy2
           y2 = y1 - czy2;
           t2 = t1 + czy2;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t2 * width + x1;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {
             if (pathzy2[iz].ind <= czy2) {
                ck = 1;
                y2 = y1 - (iz + 1);
                t2 = t1 + iz + 1;
                edges[num].w = diff2(im, x1, t1, y1, x1, t2, y2, n1, n2, e1, e2); 
             }
            }
           num++; 
          }       

         if (czx2 > 0) { // Do pathzx2
           y2 = y1 - czx2;
           x2 = x1 + czx2;
           edges[num].a = y1 * hw + t1 * width + x1;
           edges[num].b = y2 * hw + t1 * width + x2;

           ck = 0;
           for (int iz=0; ck == 0; iz++)  {
             if (pathzx2[iz].ind <= czx2) {
                ck = 1;
                y2 = y1 - (iz + 1);
                x2 = t1 + iz + 1;
                edges[num].w = diff2(im, x1, t1, y1, x2, t1, y2, n1, n2, e1, e2); 
             }
            }
           num++;        
          }
    }
  }
 } 


fprintf(stderr,"edges done: %d \n", num);

  // segment
  universe *u = segment_graph(width*height*depth, num, edges, c);
 fprintf(stderr,"After segment \n"); 
  // post process small components
  for (int i = 0; i < num; i++) {
    int a = u->find(edges[i].a);
    int b = u->find(edges[i].b);
    if ((a != b) && ((u->size(a) < min_size) || (u->size(b) < min_size)))
      u->join(a, b);
  }
  delete [] edges;
  
  
  *num_ccs = u->num_sets();
  float *colors = new float[n1*n2*n3];
   srand( time(NULL) );
  for (int q = 0; q < n1*n2*n3; q++) { 
    colors[q] = 10*(float(rand()) / (RAND_MAX+1) + 1)+5*(float(rand()) / (RAND_MAX+1) + 1);

}
  output = (float*)malloc(n1*n2*n3*4);

  

  //float *output = new image<float>(width, height);

  // pick random colors for each component
 // float *colors = new float[width*height];
 // for (int i = 0; i < width*height; i++) {
   // colors[i] = (int)rand();  //random_rgb();
    //fprintf(stderr,"color: %d \n",colors[i]);
//}
 for (int y = 0; y < depth; y++) { 
  for (int t = 0; t < height; t++) {
    for (int x = 0; x < width; x++) {
      int comp = u->find(y*hw + t * width + x);        
      output[y*n2*n1 + t + x*n1] = colors[comp];
    }
  } 
}

/*
  for (int z = 0; z < depth; z++) { 

   for (int y = 1; y < height; y++) {
    for (int x = 1; x < width; x++) {
 
      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y-1]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y+1]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + x*n1+y-1];

      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x-1)*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x+1)*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + (x-1)*n1+y];
/*
      if (( z > 0 ) && (z < depth-1) ) {
        if ( (output[z*n1*n2 + x*n1+y] != output[(z-1)*n1*n2 + x*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[(z+1)*n1*n2 + x*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[(z-1)*n1*n2 + x*n1+y];

      }

    }
  } 
}

  for (int z = 0; z < depth; z++) { 
   for (int y = 1; y < height; y++) {
    for (int x = 1; x < width; x++) {
 
      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y-1]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y+1]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + x*n1+y-1];

      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x-1)*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x+1)*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + (x-1)*n1+y];
/*
      if ( z > 0 ) {
        if ( (output[z*n1*n2 + x*n1+y] != output[(z-1)*n1*n2 + x*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[(z+1)*n1*n2 + x*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[(z-1)*n1*n2 + x*n1+y];

      }

    }
  } 
}
*/
  delete [] colors;  
  delete u;

  return output;
}

#endif


