/*
Modified from code copyright (C) 2006 Pedro Felzenszwalb, under the terms of the GNU General Public License as published by
the Free Software Foundation.


*/

#ifndef SEGMENT_IMAGE
#define SEGMENT_IMAGE

#include <cstdlib>
//#include <image.h>
#include <algorithm>
#include <cmath>
#include "segment-graph.h"

typedef struct {
  float val;
  int ind, loc;
} ind_val;

bool operator<(const ind_val &a, const ind_val &b) {
  return a.val > b.val;
}

// dissimilarity measure between pixels

static inline float diff2(float *r, int x1, int y1, int x2, int y2, int n1, int e1, int e2) {
  
  float dist=sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2)) / e1;
  float amp=r[y2+n1*x2] / e2;
  return exp((dist+amp));
}

/*
 * Segment an image
 *
 * Returns a color image representing the segmentation.
 *
 * im: image to segment.
 * sigma: to smooth the image.
 * c: constant for treshold function.
 * min_size: minimum component size (enforced by post-processing stage).
 * num_ccs: number of connected components in the segmentation.
 */

float *segment_image(float *im, int n1, int n2, float sigma, float c, int min_size, int e1, int e2, int max_dist, int *num_ccs) {
  int width = n2, ck, ckx, x1, x2, t1, t2;
  int height = n1;
  float *output; 
  

fprintf(stderr,"in segment_image: %d x %d \n",width,height); 



  // build graph 
  //int max_dist = 6;
  edge *edges = new edge[width*height*max_dist*4];


  
  int num = 0;
  for (int t1 = 0; t1 < height; t1++) {
    for (int x1 = 0; x1 < width; x1++) {

    int cx=0, cy=0, cxy=0, cxy2=0;

 
   ind_val *pathx = new ind_val[max_dist];
   ind_val *pathy = new ind_val[max_dist];
   ind_val *pathxy = new ind_val[max_dist];
   ind_val *pathxy2 = new ind_val[max_dist];

       for (int px=0; px < max_dist; px++) {
          int x2 = x1 + px +1;
          int t2 = t1 + px +1;
          int t3 = t1 - (px+1);
          if (x2 < width) {
             pathx[px].ind = px;
             pathx[px].val = im[t1 * width + x2];
             pathx[px].loc = t1 * width + x2;
             cx++;
           }
          if (t2 < height) {
             pathy[px].ind = px;
             pathy[px].val = im[t2 * width + x1];
             pathy[px].loc = t2 * width + x1;
             cy++;
           }
         if ((x2 < width) && (t2 < height)) {
             pathxy[px].ind = px;
             pathxy[px].val = im[t2 * width + x2];
             pathxy[px].loc = t2 * width + x2;
             cxy++;
           }
         if ((x2 < width) && (t3 > 0) && (t2 < height)) {
             pathxy2[px].ind = px;
             pathxy2[px].val = im[t3 * width + x2];
             pathxy2[px].loc = t3 * width + x2;                        
             cxy2++;
           }


        std::sort(pathx, pathx + cx);
        std::sort(pathy, pathy + cy);
        std::sort(pathxy, pathxy + cxy);
        std::sort(pathxy2, pathxy2 + cxy2);
        
         }

         for (int px=0; px < cx; px++) {
           x2 = x1 + px +1;
           edges[num].a = t1 * width + x1;
           edges[num].b = t1 * width + x2;
           ck = 0;
           for (int ix=0; ck == 0; ix++)  {
             if (pathx[ix].ind <= px) {
                ck = 1;
                ckx = pathx[ix].ind;
                x2 = x1 + ix + 1;
                edges[num].w = diff2(im, x1, t1, x2, t1, n1, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
             }
            }       
           num++;        
          } 


       for (int pt=0; pt < cy; pt++) {
          int t2 = t1 + pt +1;
           edges[num].a = t1 * width + x1;
           edges[num].b = t2 * width + x1;

           ck = 0;
           for (int iy=0; ck == 0; iy++)  {
             if (pathy[iy].ind <= pt) {
                ck = 1;
                t2 = t1 + iy + 1;
                edges[num].w = diff2(im, x1, t1, x1, t2, n1, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathy[iy].loc]));
             }
            }
          num++;        
         }

       for (int pd=0; pd < cxy; pd++) {
          int x2 = x1 + pd +1;
          int t2 = t1 + pd +1;
           edges[num].a = t1 * width + x1;
           edges[num].b = t2 * width + x2;

           ck = 0;
           for (int id=0; ck == 0; id++)  {
             if (pathxy[id].ind <= pd) {
                ck = 1;
                x2 = x1 + id + 1;
                t2 = t1 + id + 1;
                edges[num].w = diff2(im, x1, t1, x2, t2, n1, e1, e2);  //sqrt(square(im[t1 * width + x1]-im[pathxy[id].loc]));
             }
            }
           num++;
         }

       for (int pd=0; pd < cxy2; pd++) {
          int x2 = x1 + pd +1;
          int t2 = t1 - (pd+1);
           edges[num].a = t1 * width + x1;
           edges[num].b = t2 * width + x2;
           ck = 0;
           for (int id=0; ck == 0; id++)  {
             if (pathxy2[id].ind <= pd) {
                ck = 1;
                x2 = x1 + id + 1;
                t2 = t1 - (id + 1);
                edges[num].w = diff2(im, x1, t1, x2, t2, n1, e1, e2);  //sqrt(square(im[t1 * width + x1]-im[pathxy2[id].loc]));
             }
            }
           num++;
         }

    }
  }

fprintf(stderr,"edges done: %d \n", num);

  // segment
  universe *u = segment_graph(width*height, num, edges, c);
 fprintf(stderr,"After segment"); 
  // post process small components
  for (int i = 0; i < num; i++) {
    int a = u->find(edges[i].a);
    int b = u->find(edges[i].b);
    if ((a != b) && ((u->size(a) < min_size) || (u->size(b) < min_size)))
      u->join(a, b);
  }
  delete [] edges;
  *num_ccs = u->num_sets();
  float *colors = new float[n1*n2];
   srand( time(NULL) );
  for (int q = 0; q < n1*n2; q++) { 
    colors[q] = 10*(float(rand()) / (RAND_MAX+1) + 1)+5*(float(rand()) / (RAND_MAX+1) + 1);
    // fprintf(stderr,"color %d \n",colors[q]);
}
  output = (float*)malloc(n1*n2*4);


  // pick random colors for each component
 // float *colors = new float[width*height];
 // for (int i = 0; i < width*height; i++) {
   // colors[i] = (int)rand();  //random_rgb();
    //fprintf(stderr,"color: %d \n",colors[i]);
//}
  
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      int comp = u->find(y * width + x);
         
      output[x*n1+y] = colors[comp];
    }
  } 

   for (int y = 1; y < height; y++) {
    for (int x = 1; x < width; x++) {
 
      if ( (output[x*n1+y] != output[x*n1+y-1]) && (output[x*n1+y] != output[x*n1+y+1]))
            output[x*n1+y] = output[x*n1+y-1];

      if ( (output[x*n1+y] != output[(x-1)*n1+y]) && (output[x*n1+y] != output[(x+1)*n1+y]))
            output[x*n1+y] = output[(x-1)*n1+y];

    }
  } 

   for (int y = 1; y < height; y++) {
    for (int x = 1; x < width; x++) {
 
      if ( (output[x*n1+y] != output[x*n1+y-1]) && (output[x*n1+y] != output[x*n1+y+1]))
            output[x*n1+y] = output[x*n1+y-1];

      if ( (output[x*n1+y] != output[(x-1)*n1+y]) && (output[x*n1+y] != output[(x+1)*n1+y]))
            output[x*n1+y] = output[(x-1)*n1+y];

    }
  } 



  delete [] colors;  
  delete u;

  return output;
}

#endif


