/*

Modified from code copyright (C) 2006 Pedro Felzenszwalb, under the terms of the GNU General Public License as published by
the Free Software Foundation.

*/

#ifndef SEGMENT_IMAGE
#define SEGMENT_IMAGE

#include <cstdlib>
//#include <image.h>
#include <algorithm>
#include <cmath>
#include "segment-graph.h"

typedef struct {
  float val;
  int ind, loc;
} ind_val;

bool operator<(const ind_val &a, const ind_val &b) {
  return a.val > b.val;
}
 static inline float diff2(float *r, int x1, int t1, int y1, int x2, int t2, int y2, int n1, int n2, float e1, float e2) {
  
  return sqrt((r[t2+n1*x2+y2*n1*n2]-r[t1+n1*x1+y1*n1*n2])*(r[t2+n1*x2+y2*n1*n2]-r[t1+n1*x1+y1*n1*n2]));
}


/*
 * Segment an image
 *
 * Returns a color image representing the segmentation.
 *
 * im: image to segment.
 * sigma: to smooth the image.
 * c: constant for treshold function.
 * min_size: minimum component size (enforced by post-processing stage).
 * num_ccs: number of connected components in the segmentation.
 */
//image<rgb> *segment_image(image<rgb> *im, float sigma, float c, int min_size, int *num_ccs) {

float *segment_image(float *im, int n1, int n2, int n3, float sigma, float c, int min_size, int max_dist, int *num_ccs) {
  int width = n2, ck, ckx, x1, x2, t1, t2;
  int height = n1;
  int depth = n3;
  float *output; 
  

fprintf(stderr,"in segment_image: %d x %d \n",width,height);

  // build graph

  float e1 = 1; float e2=e1;
  edge *edges = new edge[width*height*depth*15];


  int num = 0;
 for (int y1 = 0; y1 < depth; y1++)  {
fprintf(stderr,"slice %d \n",y1);
  for (int t1 = 0; t1 < height; t1++) {
    for (int x1 = 0; x1 < width; x1++) {

   
         if (x1 < width-1) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = y1 * height*width + t1 * width + x1+1;
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1, y1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          } 

         if (t1 < height-1) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = y1 * height*width + (t1+1) * width + x1;
           edges[num].w = diff2(im, x1, t1, y1, x1, t1+1, y1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          } 
          
         if (y1 < depth-1) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1+1) * height*width + t1 * width + x1;
           edges[num].w = diff2(im, x1, t1, y1, x1, t1, y1+1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          }
          
         if ((x1 < width-1) && (t1 < height-1)) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = y1 * height*width + (t1+1) * width + (x1+1);
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1+1, y1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          }   

         if ((x1 < width-1) && (t1 > 0)) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = y1 * height*width + (t1-1) * width + (x1+1);
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1-1, y1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          } 
          

         if ((y1 < depth-1) && (t1 < height-1) && (x1<width-1)) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1+1) * height*width + (t1+1) * width + (x1+1);
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1+1, y1+1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          }          
          
         if ((y1 < depth-1) && (t1 > 0) && (x1<width-1)) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1+1) * height*width + (t1-1) * width + (x1+1);
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1-1, y1+1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          }    
          
         if ((y1 < depth-1) && (t1 < height-1)) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1+1) * height*width + (t1+1) * width + (x1);
           edges[num].w = diff2(im, x1, t1, y1, x1, t1+1, y1+1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          } 
          
         if ((y1 < depth-1) && (x1<width-1)) {
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1+1) * height*width + (t1) * width + (x1+1);
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1, y1+1, n1, n2, e1, e2); //sqrt(square(im[t1 * width + x1]-im[pathx[ix].loc]));
           num++;        
          }  

         if ((x1 < width-1) && (y1 > 0) && (t1 < height-1)) {        
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1-1) * height*width + (t1+1) * width + (x1+1);
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1+1, y1-1, n1, n2, e1, e2);
           num++;        
          } 

         if ((x1 < width-1) && (y1 > 0) ) {        
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1-1) * height*width + (t1) * width + (x1+1);
           edges[num].w = diff2(im, x1, t1, y1, x1+1, t1, y1-1, n1, n2, e1, e2);
           num++;        
          } 

         if ((t1 < height-1) && (y1 > 0) ) {        
           edges[num].a = y1 * height*width + t1 * width + x1;
           edges[num].b = (y1-1) * height*width + (t1+1) * width + x1;
           edges[num].w = diff2(im, x1, t1, y1, x1, t1+1, y1-1, n1, n2, e1, e2);
           num++;        
          } 
    }
  }
 } 


fprintf(stderr,"edges done: %d \n", num);

  // segment
  universe *u = segment_graph(width*height*depth, num, edges, c);
 fprintf(stderr,"After segment \n"); 
  // post process small components
  for (int i = 0; i < num; i++) {
    int a = u->find(edges[i].a);
    int b = u->find(edges[i].b);
    if ((a != b) && ((u->size(a) < min_size) || (u->size(b) < min_size)))
      u->join(a, b);
  }
  delete [] edges;
  *num_ccs = u->num_sets();
  float *colors = new float[n1*n2*n3];
   srand( time(NULL) );
  for (int q = 0; q < n1*n2*n3; q++) { 
    colors[q] = 10*(float(rand()) / (RAND_MAX+1) + 1)+5*(float(rand()) / (RAND_MAX+1) + 1);

}
  output = (float*)malloc(n1*n2*n3*4);

  //float *output = new image<float>(width, height);

  // pick random colors for each component
 // float *colors = new float[width*height];
 // for (int i = 0; i < width*height; i++) {
   // colors[i] = (int)rand();  //random_rgb();
    //fprintf(stderr,"color: %d \n",colors[i]);
//}
 for (int y = 0; y < depth; y++) { 
  for (int t = 0; t < height; t++) {
    for (int x = 0; x < width; x++) {
      int comp = u->find(y*height*width + t * width + x);

      output[y*n2*n1 + t + x*n1] = colors[comp];
    }
  } 
}

/*
  for (int z = 0; z < depth; z++) { 

   for (int y = 1; y < height; y++) {
    for (int x = 1; x < width; x++) {
 
      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y-1]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y+1]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + x*n1+y-1];

      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x-1)*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x+1)*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + (x-1)*n1+y];
/*
      if (( z > 0 ) && (z < depth-1) ) {
        if ( (output[z*n1*n2 + x*n1+y] != output[(z-1)*n1*n2 + x*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[(z+1)*n1*n2 + x*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[(z-1)*n1*n2 + x*n1+y];

      }

    }
  } 
}

  for (int z = 0; z < depth; z++) { 
   for (int y = 1; y < height; y++) {
    for (int x = 1; x < width; x++) {
 
      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y-1]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + x*n1+y+1]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + x*n1+y-1];

      if ( (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x-1)*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[z*n1*n2 + (x+1)*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[z*n1*n2 + (x-1)*n1+y];
/*
      if ( z > 0 ) {
        if ( (output[z*n1*n2 + x*n1+y] != output[(z-1)*n1*n2 + x*n1+y]) && (output[z*n1*n2 + x*n1+y] != output[(z+1)*n1*n2 + x*n1+y]))
            output[z*n1*n2 + x*n1+y] = output[(z-1)*n1*n2 + x*n1+y];

      }

    }
  } 
}

*/

  delete [] colors;  
  delete u;

  return output;
}

#endif


