/*


Modified from code copyright (C) 2006 Pedro Felzenszwalb, under the terms of the GNU General Public License as published by
the Free Software Foundation.


Segments a 2D seismic image (SEP format) using the Pairwise Region Comparison 
 method of Felzenszwalb and Huttenlocher (2004).  Searches for boundaries in the
 input data, which should be amplitude of the envelope or similar.
 
 Usage:
 < input_envelope.H segment.x k min_size e1 e2 max_dist > output.H
 	k: scaling parameter; typically small (~0.001)
 	min_size: minimum output segment size, in pixels
 	e1, e2: Distance weighting parameters; recommend both at 1
 	max_dist: length (in pixels) of stencil arms when constructing graph

*/

#include <cstdio>
#include <cstdlib>
#include "segment-image.h"
extern "C" { 
  #include "sep3d.h"
}

int main(int argc, char **argv) { 

sep_begin_prog();
initpar(argc, argv);
init_3d();

fprintf(stderr,"argc=%d \n",argc);
 for (int iarg=1; iarg<argc; iarg++) { 
   getch_add_string(argv[iarg]);  
 }

int n1,n2,min_size,e1=1,e2=1,max_dist=5;
float sigma=.5,k=.0001;
float *data;

if (!hetch ("n1","d", &n1))
      seperr("no n1");
if (!hetch ("n2","d", &n2))
      seperr("no n2");

if (!fetch ("k","f", &k))
      fprintf(stderr,"using k=%f \n",k);
if (!fetch ("min_size","d", &min_size))
      seperr("Did not get min_size \n");
if (!fetch ("max_dist","d", &max_dist))
      fprintf(stderr,"using max_dist=%d \n",max_dist);


data = (float*)malloc(n1*n2*4);
if (!sreed("in",data,n1*n2*4))
   fprintf(stderr,"Problem reading in");


fprintf(stderr,"n1 is %d, n2 is %d, sigma is %f, k is %f, min_size is %d \n", n1,n2,sigma,k,min_size);
fprintf(stderr,"argc is %d, e1 is %d, e2 is %d \n", argc, e1, e2);

  fprintf(stderr,"processing\n");
  int num_ccs; 
  float *seg = segment_image(data, n1, n2, sigma, k, min_size, e1, e2, max_dist, &num_ccs);

  fprintf(stderr,"got %d components\n", num_ccs);

 if (n1*n2*4!=srite("out",seg,n1*n2*4))
     seperr("Trouble writing out"); 
  
sep_end_prog();
return 0;

}

