/*
Modified from code copyright (C) 2006 Pedro Felzenszwalb, under the terms of the GNU General Public License as published by
the Free Software Foundation.


Segments a 3D seismic image (SEP format) using the Pairwise Region Comparison 
 method of Felzenszwalb and Huttenlocher (2004).
 
 Usage:
 < input_envelope.H segment.x k min_size e1 e2 max_dist > output.H
 	k: scaling parameter; typically small (~0.001)
 	min_size: minimum output segment size, in pixels
 	e1, e2: Distance weighting parameters; recommend both at 1
 	max_dist: length (in pixels) of stencil arms when constructing graph
*/

#include <cstdio>
#include <cstdlib>
#include "omp.h"
#include "segment-image_3d.h"
extern "C" {  
 #include "sep3d.h"
    }

main(int argc, char **argv) { 

sep_begin_prog();
initpar(argc, argv);
init_3d();

fprintf(stderr,"argc=%d \n",argc);
 for (int iarg=1; iarg<argc; iarg++) { 
   getch_add_string(argv[iarg]);  
 }

int n1,n2,n3,n4,np,min_size,max_dist=5;
float sigma=0,k=.0001,e1=1,e2=1;
float *data, *int_bnd, *seg;
if (!hetch ("n1","d", &n1))
      seperr("no n1");
if (!hetch ("n2","d", &n2))
      seperr("no n2");
if (!hetch ("n3","d", &n3))
      seperr("no n3");
if (!hetch ("n4","d", &n4))
      seperr("no n3");

if (!fetch ("k","f", &k))
      fprintf(stderr,"using k=%f \n",k);
if (!fetch ("min_size","d", &min_size))
      fprintf(stderr,"Did not get min_size \n");
if (!fetch ("max_dist","d", &max_dist))
      fprintf(stderr,"using max_dist=%d \n",max_dist);

double memsz=n1*n2*n3*n4*4;
data = (float*)malloc(memsz);
seg = (float*)malloc(memsz);
if (!sreed("in",data,n1*n2*n3*n4*4))
   fprintf(stderr,"Problem reading in");

fprintf(stderr,"Doing %d patches \n",n4);
fprintf(stderr,"n1 is %d, n2 is %d, n3 is %d, k is %f, min_size is %d, max_dist is %d \n", n1,n2,n3,k,min_size,max_dist);
fprintf(stderr,"argc is %d, e1 is %f, e2 is %f \n", argc, e1, e2);

  fprintf(stderr,"processing\n");



  #pragma omp parallel //default(shared) private(pseg,patch,num_ccs,i,j)  
  {
   int nth=omp_get_num_threads();
   int ith=omp_get_thread_num();
   int num_ccs,i,j;
   float *patch = (float*)malloc(n1*n2*n3*4);
   for (i=ith*n1*n2*n3,j=0;j<(n1*n2*n3);i++,j++) {
    patch[j]=data[i];
   }

   float *pseg = segment_image(patch, n1, n2, n3, sigma, k, min_size, e1, e2, max_dist, &num_ccs);

   for (i=ith*n1*n2*n3,j=0;j<(n1*n2*n3);i++,j++) {
    seg[i]=pseg[j];
   }
  }
  
 fprintf(stderr,"got %d segments\n", num_ccs);

 if (n1*n2*n3*n4*4!=srite("out",seg,n1*n2*n3*n4*4))
     seperr("Trouble writing out"); 

  
sep_end_prog();
return 0;

}

